!> Simple framework for error handling in Fortran.
!
!  This error handling strategy has been described in:
!
!  Poppe, Cools,Vandewoestyne: 
!  Error handling in Fortran 2003 (2012)
!
!  It benefits from a number of F2003 features.  In this
!  implementation, some parts of the framework have been simplified.
!
!  Implementation: 09/2016, F. Prill (DWD)
!
MODULE error_handling

  IMPLICIT NONE

  PUBLIC :: t_error
  PUBLIC :: t_error_info, t_invalid_input_error, t_arithmetic_error

  PUBLIC :: create_error
  PUBLIC :: transfer_error

  PUBLIC :: MAX_ERROR_STR_LEN

  INTEGER, PARAMETER :: MAX_ERROR_STR_LEN = 256

  !> Base class for error info, storing the details of the
  !  circumstances in which the error occured. Making a distinction
  !  between the error and the associated information simplifies the
  !  use.
  TYPE t_error_info   
    CHARACTER(LEN=MAX_ERROR_STR_LEN) :: error_string
  END TYPE t_error_info

  !> Error info type: Invalid user input
  TYPE, EXTENDS(t_error_info) :: t_invalid_input_error
  END TYPE t_invalid_input_error

  !> Error info type: Arithmetic error
  TYPE, EXTENDS(t_error_info) :: t_arithmetic_error
  END TYPE t_arithmetic_error

  !> Base class for errors
  TYPE t_error
    CLASS(t_error_info), POINTER :: info => NULL()
  CONTAINS 
    PROCEDURE :: report  => error_report
    PROCEDURE :: discard => error_discard
  END TYPE t_error

CONTAINS

  !> Called in an unexpected situation, this subroutine creates an
  !  error, storing all parameters that are needed to uniquely define
  !  the situation.
  !
  SUBROUTINE create_error(new_error, new_info)
    TYPE(t_error),       INTENT(INOUT) :: new_error
    CLASS(t_error_info), INTENT(IN)    :: new_info
    INTEGER :: ierrstat
    ALLOCATE(new_error%info, source=new_info, STAT=ierrstat)
    IF (ierrstat /= 0) THEN
      WRITE (0,*) "error_discard: DEALLOCATE failed!"
    END IF
  END SUBROUTINE create_error

  !> Transfers the error from @p inform to the @p ifail argument.
  !
  !  Thus, the responsibility of the error can be transferred to the
  !  calling procedure.
  !
  SUBROUTINE transfer_error(inform, new_error)
    TYPE(t_error), INTENT(IN)    :: inform
    TYPE(t_error), INTENT(INOUT) :: new_error
    IF (ASSOCIATED(new_error%info))  CALL new_error%report()
    new_error%info => inform%info
  END SUBROUTINE transfer_error

  !> Outputs the given error.
  !
  SUBROUTINE error_report(this)
    CLASS(t_error) :: this
    IF (ASSOCIATED(this%info)) THEN
      WRITE (0,*) "ERROR STRING: ", this%info%error_string
    ELSE
      WRITE (0,*) "SUCCESS"
    END IF
    CALL this%discard()
  END SUBROUTINE error_report

  !> Indicates that the given error has been taken care of and can be
  !  silenced safely.
  !
  SUBROUTINE error_discard(this)
    CLASS(t_error) :: this
    INTEGER        :: ierrstat
    IF (ASSOCIATED(this%info)) THEN
      DEALLOCATE(this%info, STAT=ierrstat)
      IF (ierrstat /= 0) THEN
        WRITE (0,*) "error_discard: DEALLOCATE failed!"
      END IF
    END IF
  END SUBROUTINE error_discard

END MODULE error_handling
