!> @file libmtime.f90
!!
!! @brief Providing a high level interface for the Fortran language bindings of libmtime
!!
!! @author  Luis Kornblueh, Max Planck Institute for Meteorology
!! @author  Florian Prill, DWD
!!
!! @defgroup FortranBindings libmtime high level Fortran language bindings
!! @{
!!
module mtime_hl

  use mtime
  use error_handling

  implicit none
  
  private

  public :: t_datetime, t_timedelta

  INTEGER, PARAMETER :: i8 = SELECTED_INT_KIND(14)   !< at least 8 byte integer

  !> Wrapper class for "mtime" data type "datetime".
  !
  !  Avoids, e.g.,  explicit use of POINTERs by the user.
  !
  type t_datetime

    private

    type(datetime) :: dt

  contains

    procedure :: day                       => t_datetime_day
    procedure :: assign_string             => t_datetime_assign_string
    procedure :: add_timedelta             => t_datetime_add_timedelta
    procedure :: sub_timedelta             => t_datetime_sub_timedelta    
    procedure :: sub_datetime              => t_datetime_sub_datetime
    procedure :: equal_datetime            => t_datetime_equal
    procedure :: not_equal_datetime        => t_datetime_not_equal
    procedure :: less_than_datetime        => t_datetime_less_than
    procedure :: greater_than_datetime     => t_datetime_greater_than
    procedure :: less_or_equal_datetime    => t_datetime_less_or_equal
    procedure :: greater_or_equal_datetime => t_datetime_greater_or_equal

    procedure :: t_datetime_to_string
    procedure :: t_datetime_to_posix_string

    generic   :: to_string => t_datetime_to_string, t_datetime_to_posix_string 
    
    generic   :: assignment(=) =>  assign_string
    generic   :: operator(+)   =>  add_timedelta
    generic   :: operator(-)   =>  sub_timedelta    
    generic   :: operator(-)   =>  sub_datetime
    generic   :: operator(==)  =>  equal_datetime
    generic   :: operator(/=)  =>  not_equal_datetime       
    generic   :: operator(<)   =>  less_than_datetime       
    generic   :: operator(>)   =>  greater_than_datetime    
    generic   :: operator(<=)  =>  less_or_equal_datetime   
    generic   :: operator(>=)  =>  greater_or_equal_datetime 
    
  end type t_datetime


  !> Wrapper class for "mtime" data type "timedelta".
  !
  !  Avoids, e.g.,  explicit use of POINTERs by the user.
  !
  type t_timedelta

    private

    type(timedelta) :: td

  contains

    procedure :: assign_string    => t_timedelta_assign_string
    procedure :: assign_ms        => t_timedelta_assign_ms

    procedure :: t_timedelta_to_string
    
    generic   :: to_string        => t_timedelta_to_string
    
    generic   :: assignment(=)   => assign_string, assign_ms

  end type t_timedelta


  !> Error info type: mtime error
  type, extends(t_error_info) :: t_mtime_error
  end type t_mtime_error

contains 

  ! Returns t_datetime objects day
  !
  function t_datetime_day(this)
    integer :: t_datetime_day
    class (t_datetime)                   :: this
    t_datetime_day = this%dt%date%day
  end function t_datetime_day

  ! Convert t_datetime object to string.
  !
  function t_datetime_to_string(this, opt_error)
    character(len=max_datetime_str_len)  :: t_datetime_to_string
    type(t_error), intent(out), optional :: opt_error
    class (t_datetime)                   :: this
    type(datetime), pointer              :: dt_tmp
    integer                              :: errno
    character(len=max_error_str_len)     :: error_str
    dt_tmp => newdatetime(this%dt)
    if (present(opt_error)) then
      call datetimetostring(dt_tmp, t_datetime_to_string, errno)    
      if  (errno /= 0) then
        call mtime_strerror(errno, error_str)
        call create_error(opt_error, t_mtime_error(error_str))      
      end if
    else
      call datetimetostring(dt_tmp, t_datetime_to_string)    
    end if
    call deallocatedatetime(dt_tmp)
  end function t_datetime_to_string

   ! Convert t_datetime object to string.
  !
  function t_datetime_to_posix_string(this, format_string, opt_error)
    character(len=max_datetime_str_len)  :: t_datetime_to_posix_string
    character(len=*), intent(in)         :: format_string
    type(t_error), intent(out), optional :: opt_error
    class (t_datetime)                   :: this
    type(datetime), pointer              :: dt_tmp
    integer                              :: errno
    character(len=max_error_str_len)     :: error_str
    dt_tmp => newdatetime(this%dt)
    if (present(opt_error)) then
      call datetimetoposixstring(dt_tmp, t_datetime_to_posix_string, format_string, errno)    
      if  (errno /= 0) then
        call mtime_strerror(errno, error_str)
        call create_error(opt_error, t_mtime_error(error_str))      
      end if
    else
      call datetimetostring(dt_tmp, t_datetime_to_posix_string)    
    end if
    call deallocatedatetime(dt_tmp)
  end function t_datetime_to_posix_string
 
  ! Assignment operator: this = "date string".
  !
  subroutine t_datetime_assign_string(this, dt_string)
    class (t_datetime), intent(out) :: this
    character(len=*),   intent(in)  :: dt_string
    type(datetime), pointer :: dt_tmp
    dt_tmp => newDatetime(dt_string)
    this%dt = dt_tmp
    call deallocateDatetime(dt_tmp)
  end subroutine t_datetime_assign_string

  ! Addition of time interval to datetime object.
  !
  function t_datetime_add_timedelta(this, td) result(dt_td_sum)
    type (t_datetime)               :: dt_td_sum
    class (t_datetime),  intent(in) :: this    
    class (t_timedelta), intent(in) :: td
    type(datetime),  pointer :: dt_tmp
    type(timedelta), pointer :: td_tmp
    dt_tmp => newDatetime(this%dt)
    td_tmp => newTimedelta(td%td)
    dt_tmp = dt_tmp + td_tmp
    dt_td_sum%dt = dt_tmp
    call deallocateDatetime(dt_tmp)
    call deallocateTimedelta(td_tmp)
  end function t_datetime_add_timedelta

  ! Subtraction of time interval to datetime object.
  !
  function t_datetime_sub_timedelta(this, td) result(dt_td_sum)
    type (t_datetime)               :: dt_td_sum
    class (t_datetime),  intent(in) :: this    
    class (t_timedelta), intent(in) :: td
    type(datetime),  pointer :: dt_tmp
    type(timedelta), pointer :: td_tmp
    dt_tmp => newDatetime(this%dt)
    td_tmp => newTimedelta(td%td)
    if (td_tmp%sign == "+") then
      td_tmp%sign = "-" 
    else
      td_tmp%sign = "+"       
    endif
    dt_tmp = dt_tmp + td_tmp
    dt_td_sum%dt = dt_tmp
    call deallocateDatetime(dt_tmp)
    call deallocateTimedelta(td_tmp)
  end function t_datetime_sub_timedelta

  ! Subtraction of two dates.
  !
  function t_datetime_sub_datetime(this, dt) result(dt_dt_diff)
    type (t_timedelta)              :: dt_dt_diff
    class (t_datetime),  intent(in) :: this
    class (t_datetime),  intent(in) :: dt
    dt_dt_diff%td = getTimeDeltaFromDateTime(this%dt, dt%dt)
  end function t_datetime_sub_datetime

  ! Overloaded operator: test for equivalence.
  !
  logical function t_datetime_equal(this, dt) 
    class (t_datetime),  intent(in) :: this    
    class (t_datetime),  intent(in) :: dt
    t_datetime_equal = (this%dt == dt%dt)
  end function t_datetime_equal

  logical function t_datetime_not_equal(this, dt)
    class (t_datetime),  intent(in) :: this    
    class (t_datetime),  intent(in) :: dt
    t_datetime_not_equal = (this%dt /= dt%dt)
  end function t_datetime_not_equal

  logical function t_datetime_less_than(this, dt) 
    class (t_datetime),  intent(in) :: this    
    class (t_datetime),  intent(in) :: dt
    t_datetime_less_than = (this%dt < dt%dt)
  end function t_datetime_less_than

  logical function t_datetime_greater_than(this, dt) 
    class (t_datetime),  intent(in) :: this    
    class (t_datetime),  intent(in) :: dt
    t_datetime_greater_than = (this%dt > dt%dt)
  end function t_datetime_greater_than

  logical function t_datetime_less_or_equal(this, dt) 
    class (t_datetime),  intent(in) :: this    
    class (t_datetime),  intent(in) :: dt
    t_datetime_less_or_equal = (this%dt <= dt%dt)
  end function t_datetime_less_or_equal

  logical function t_datetime_greater_or_equal(this, dt) 
    class (t_datetime),  intent(in) :: this    
    class (t_datetime),  intent(in) :: dt
    t_datetime_greater_or_equal = (this%dt >= dt%dt)
  end function t_datetime_greater_or_equal
  
  ! Convert t_timedelta object to string.
  !
  FUNCTION t_timedelta_to_string(this, opt_error)
    character(len=max_timedelta_str_len) :: t_timedelta_to_string
    class (t_timedelta)                  :: this
    TYPE(t_error), INTENT(OUT), OPTIONAL :: opt_error
    type(timedelta), pointer             :: td_tmp
    INTEGER                              :: errno
    character(len=max_error_str_len)     :: error_str
    td_tmp => newtimedelta(this%td)
    if (present(opt_error)) then
      call timedeltatostring(td_tmp, t_timedelta_to_string, errno)    
      if  (errno /= 0) then
        call mtime_strerror(errno, error_str)
        call create_error(opt_error, t_mtime_error(error_str))      
      end if
    else
      call timedeltatostring(td_tmp, t_timedelta_to_string)    
    end if
    call deallocatetimedelta(td_tmp)
  end function t_timedelta_to_string

  ! Assignment operator: this = "time interval string".
  !
  subroutine t_timedelta_assign_string(this, td_string)
    class (t_timedelta), intent(out) :: this
    character(len=*),    intent(in)  :: td_string
    type(timedelta), pointer :: td_tmp
    td_tmp => newtimedelta(td_string)
    this%td = td_tmp
    this%td%sign = td_tmp%sign
    call deallocatetimedelta(td_tmp)
  end subroutine t_timedelta_assign_string

  ! Assignment operator: this = milliseconds.
  !
  subroutine t_timedelta_assign_ms(this, td_ms)
    class (t_timedelta), intent(out) :: this
    integer,             intent(in)  :: td_ms
    type(timedelta), pointer :: td_tmp
    character(len=max_timedelta_str_len) :: td_string
    call getptstringfromms(int(td_ms,i8), td_string)
    td_tmp => newtimedelta(td_string)
    this%td = td_tmp
    this%td%sign = td_tmp%sign
    call deallocatetimedelta(td_tmp)
  end subroutine t_timedelta_assign_ms

end module mtime_hl
!>
!! @}
